/* ===========================================================
 * OpenCA OCSPD Server - src/core.c
 * (c) 2001-2008 by Massimiliano Pala and OpenCA Labs
 * All Rights Reserved
 * ===========================================================
 * OpenCA Licensed Software
 * ===========================================================
 */

#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "core.h"
#include "sock.h"

#include "threads.h"
#include "support.h"
#include "crl.h"

extern void auto_crl_check( int );
extern OCSPD_CONFIG *ocspd_conf;

pthread_rwlock_t crl_lock = PTHREAD_RWLOCK_INITIALIZER;
/* pthread_cond_t crl_cond; */

pthread_mutex_t clifd_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t clifd_cond;

pthread_mutex_t threads_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t threads_cond;

/* Local Functions */
void handle_sigabrt ( int i );

/* Function Bodies */

int start_threaded_server ( char * bind_s, char * port_s, int queue_len,
			int nthreads, OCSPD_CONFIG * ocspd_conf ) {
	int i;
	int cycle = -1;

	struct sockaddr_in cliaddr;
	socklen_t cliaddrlen;

	struct sigaction sa;

	/* Just print a nice log message when exits */
	atexit(close_server);

	if((ocspd_conf->clifd = (int *) 
			calloc( nthreads, sizeof(int))) == NULL ) {
		fprintf(stderr, "Memory Init error!\n");
		exit(68);
	}

	if ((ocspd_conf->listenfd = Listen(bind_s, atoi(port_s),
					queue_len)) == 0) {
		syslog(LOG_ERR, "ERROR::BIND::Can not bind to [%s],[%d]",
			bind_s, atoi(port_s));
		exit(101);
	}

	/* Now Chroot the application */
	if( set_chroot( ocspd_conf ) < 1 ) {
		syslog( LOG_ERR, "ERROR: can not chroot, exiting!");
		exit(204);
	}

	/* Set privileges */
	if( set_privileges( ocspd_conf ) < 1 ) {
		if( ocspd_conf->chroot_dir != NULL ) {
			syslog( LOG_ERR, "ERROR::SECURITY:: "
				"Can not drop privileges");
			syslog( LOG_ERR, "ERROR::SECURITY:: "
				"Continuing because chrooted");
		} else {
			syslog( LOG_ERR, "FATAL ERROR::SECURITY:: "
				"Can not drop privileges!");

			exit(203);
		}
	}

	if( ocspd_conf->listenfd == 0 ) {
		fprintf(stderr, "LISTEN PARAM -> ERROR!\n");
		exit(66);
	}

	ocspd_conf->nthreads = nthreads;
	if((ocspd_conf->threads_list = calloc ( nthreads, 
					sizeof(Thread))) == NULL ) {
		fprintf(stderr, "%s:%d::ERROR::Memory allocation failed!\n",
				__FILE__, __LINE__);
	};

	ocspd_conf->iget = ocspd_conf->iput = 0;

	for( i = 0; i < nthreads; i++ ) {
		/* Initialize the values of clifd[] to -1 */
		ocspd_conf->clifd[i] = -1;

		if(thread_make(i) != 0 ) {
			syslog(LOG_ERR,
				"ERROR::FATAL::Can not create "
				"thread (%d)\n", i );
			exit(80);
		}
	}

	/* Register the alarm handler */
	set_alrm_handler();

	/* Just print a nice log message when killed */
	signal(SIGTERM, handle_sigterm );
	signal(SIGABRT, handle_sigabrt );

	/* Setting the SIGHUP in order to reload the CRLs */
	sa.sa_handler = auto_crl_check;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;

	if (sigaction(SIGHUP, &sa, NULL) == -1) {
		perror("ERROR::SIGNAL::Error during setting sig_handler");
		exit(1);
	}

	cliaddrlen = sizeof( cliaddr );

	for ( ; ; ) {
		int connfd = -1;
		int workers_free = 0;

		if( ocspd_conf->debug )
			syslog( LOG_INFO, "INFO::CORE::Waiting on connect" );

		if( (connfd = Accept ( ocspd_conf->listenfd, 
					(SA*) &cliaddr, &cliaddrlen)) == -1 ) {
			syslog( LOG_ERR, "FATAL ERROR::%s\n",
				strerror(errno));
			exit(-1);
		}
		if( ocspd_conf->verbose )
			syslog( LOG_INFO, "INFO::CORE::Connection from [%s]\n", 
	 			inet_ntoa(cliaddr.sin_addr) );

		// fprintf( stderr, "- acquiring lock clifd_mutex\n");
		/*
		if( ocspd_conf->verbose ) {
			syslog(LOG_INFO, "INFO::CORE::locking (nthreads %d)",
				ocspd_conf->nthreads );
		}
		*/

		// fprintf( stderr, "- setting iput\n");
		workers_free = 0;
		while( workers_free == 0 ) {
			cycle = (cycle + 1) % ocspd_conf->nthreads;

			if(ocspd_conf->debug ) {
				syslog(LOG_INFO, "INFO::Checking availability "
						"of Thread nr. %d [%d]", cycle,
						ocspd_conf->clifd[cycle] );
			}

			/* Lock clifd before reading */
			pthread_mutex_lock( &clifd_mutex );

			if ( ocspd_conf->clifd[cycle] == -1 ) {
				workers_free = 1;

				ocspd_conf->clifd[cycle] = connfd;

				/* Wake up all the clients */
				pthread_cond_broadcast( &threads_cond );

				/* Release the mutex */
				pthread_mutex_unlock ( &clifd_mutex );

				if(ocspd_conf->debug ) {
					syslog(LOG_INFO, "INFO::Thread %d is "
						"available::mutex released", 
						cycle );
				}

				break;
			}
			/* Just release the mutex, no clients to 
 			 * wake up */
			pthread_mutex_unlock ( &clifd_mutex );

		}

		/*
		ocspd_conf->clifd[ocspd_conf->iput] = connfd;
		ocspd_conf->iput += 1;
		*/

		/*
		if( ocspd_conf->iput == ocspd_conf->nthreads ) {
			ocspd_conf->iput = 0;
		} else if ( ocspd_conf->iput == ocspd_conf->iget ) {
			// Should never be the case!
			syslog(LOG_ERR, "ERROR::iPut [%d] == iGet [%d]",
					ocspd_conf->iput, ocspd_conf->iget);
			pthread_cond_signal( &clifd_cond );
			pthread_mutex_unlock( &clifd_mutex);
			continue;
		}

		// fprintf( stderr, "- sgnaling on the clifd_cond\n");
		pthread_cond_signal ( &clifd_cond );
		// fprintf( stderr, "- releasing the clifd_mutex\n");
		pthread_mutex_unlock ( &clifd_mutex );

		if( ocspd_conf->verbose ) {
			syslog(LOG_INFO, "INFO::CORE releasing lock "
					"(nthreads %d)", ocspd_conf->nthreads );
		}
		*/

	}
	return(0);
}

int set_alrm_handler( void ) {

	/* Now on the parent process we setup the auto_checking
	   functions */
	struct sigaction sa;

	if( ocspd_conf->crl_auto_reload ||
			ocspd_conf->crl_check_validity ) {

		int auto_rel, val_check;

		/* Help variable, for readability reasons */
		auto_rel = ocspd_conf->crl_auto_reload;
		val_check = ocspd_conf->crl_check_validity;

		/* This returns the min of the two values if it
		   is not 0, otherwise return the other */
		ocspd_conf->alarm_decrement = 
			(( auto_rel > val_check ) ? 
				(val_check ? val_check : auto_rel) : 
					(auto_rel ? auto_rel : val_check ));

		sa.sa_handler = auto_crl_check;
		sigemptyset(&sa.sa_mask);
		sa.sa_flags = SA_RESTART;

		if (sigaction(SIGALRM, &sa, NULL) == -1) {
			perror("Error during handling the death processes");
			exit(1);
		}

	 	/* signal( SIGALRM, auto_crl_check ); */
	 	alarm ( ocspd_conf->alarm_decrement );
	} else {
		signal( SIGALRM, SIG_IGN);
	}

	return 1;
}

void handle_sighup ( int i ) {

	syslog( LOG_ERR, 
		"SIGHUP::Reloading CRLs, Master!");
	ocspd_reload_crls( ocspd_conf );
	return;
}

void handle_sigterm ( int i ) {
	if( ocspd_conf->verbose ) {
		syslog( LOG_ERR, 
			"SIGTERM::Received TERM signal");
	}
	exit(0);
	return;
}

void handle_sigabrt ( int i ) {

        syslog( LOG_ERR, "ERROR::SIGABRT::received - should not happen,");
	syslog( LOG_ERR, "ERROR::SIGABRT::please enable strict locking.");
	syslog( LOG_ERR, "ERROR::SIGABRT::Fatal Error, aborting server!");

        return;
}

void close_server ( void ) {
	syslog( LOG_ERR, 
		"Exiting, Glad to serve you, Master!");
	return;
}

